#include <math.h>

/* Cumulative Normal distribution from Hull */
double cumNorm(double x){
  const double a1 = 0.319381530,
    a2 =-0.356563782,
    a3 = 1.781477937,
    a4 =-1.821255978,
    a5 = 1.330274429,
    gmma = 0.2316419;
  double k,nprime;

  if(x<0){
    return(1 - cumNorm(-x));
  }else{
    k = 1.0/(1.0 + gmma*x);
    nprime = exp(-x*x/2.0)/sqrt(2*M_PI);
    return(1 - nprime*( a1*k
                        +a2*k*k
                        +a3*k*k*k
                        +a4*k*k*k*k
                        +a5*k*k*k*k*k));
  }
}


/* Bivariate normal distribution */
double sgn(double x){
  if(x >= 0.0){
    return(1.0);
  }else{
    return(-1.0);
  }
}
/* 
  Inputs are upper limits a and b correlation rho between them.
   Formula from Hull p 260 to four decimal places 
   and 
   Z. Drezner, "Computation of Bivariate Normal Integral",
   Mathematics of Computation, vol 32, no 141, Jan 1978, pp 277-279.

   Checked with three different case from Kolb.
   Note that the following is slightly different from what one may
   infer from Hull. Where Hull may have suggested > instead of >= in
   someplaces we seem to agree with Kolb only i we have the = in place.
   See code below and also look at testbi.c for test cases.
*/
double BivarCND(double a, double b, double rho){
  int i, j;
  double A[4], B[4], sum, ap, bp, f;
  double rho1, rho2, delta;
  double cumNorm(double x);
 
  if( a * b * rho > 0){
    rho1 = sgn(a) * ( rho * a - b) / sqrt(a*a - 2.0*rho*a*b + b*b);
    rho2 = sgn(b) * ( rho * b - a) / sqrt(a*a - 2.0*rho*a*b + b*b);
    delta = (1.0 - sgn(a) * sgn(b))/4.0;
    sum = BivarCND(a,0.0,rho1) + BivarCND(b,0.0,rho2) - delta;
  }else{
    if(b <= 0 && a >= 0 && rho >= 0) {
      sum = cumNorm(b)-BivarCND(-a,b,-rho);
    }else if(a <=0 && b >=0 && rho >= 0){
      sum = cumNorm(a)-BivarCND(a,-b,-rho);
    }else if(rho <= 0.0 && a >= 0 && b >= 0){
      sum = cumNorm(a) + cumNorm(b) - 1.0 + BivarCND(-a,-b, rho); 
    }else{
      /*
        if( a <= 0.0 && b <= 0.0 && rho <=0.0 ){
      */
      /* The following if a <=0, b<=0, rho<=0 */
      A[0] = 0.3253030;
      A[1] = 0.4211071;
      A[2] = 0.1334425;
      A[3] = 0.006374323;
                  
      B[0] = 0.1337764;
      B[1] = 0.6243247;
      B[2] = 1.3425378;
      B[3] = 2.2626645;

      ap = a / sqrt(2.0 * (1.0 - rho*rho));
      bp = b / sqrt(2.0 * (1.0 - rho*rho));

      sum = 0.0;
      for(i=0;i<4;i++){
        for(j=0;j<4;j++){
          f = exp(ap * (2.0 * B[i] - ap) + bp*(2.0 * B[j] - bp) + \
                  2.0 * rho * (B[i] - ap) * (B[j] - bp));
          sum += A[i] * A[j] * f;
        }
      }
      sum *= sqrt(1.0 - rho*rho);
      sum /= M_PI;
    }
  }
  return(sum);
} /* End BivarCND */

/* Black-Scholes european call pricing with continuous dividends from Tucker
 *
 * S = current price
 * X = strike price
 * r = risk-free rate
 * s = sigma = standard deviation
 * T = periods to maturity
 * d = delta = dividend rate
 */

#define N(x) cumNorm((x));

double euroCall(double S, double X, double r, double s, double T, double d){
  double d3,Sd;
  double n1,n2;

  Sd = S*exp(-d*T);
  d3 = (log(Sd/X) + (r + (s*s/2))*T)/(s*sqrt(T));
  n1 = N(d3);
  n2 = N(d3 - s*sqrt(T));

  return( Sd*n1 - X*exp(-r*T)*n2 );
}

